==> django_pain-3.3.2-py3-none-any.whl: django_pain/models/bank.py <==
#
# Copyright (C) 2018-2024  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""Payments and invoices models."""

import uuid
from logging import getLogger
from typing import Tuple

from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import BLANK_CHOICE_DASH, CheckConstraint, Q
from django.utils.translation import gettext_lazy as _
from djmoney.models.fields import CurrencyField, MoneyField

from django_pain.constants import (
    CURRENCY_PRECISION,
    InvoiceType,
    PaymentProcessingError,
    PaymentState,
    PaymentType,
    ProcessPaymentResultType,
)
from django_pain.settings import SETTINGS, get_processor_instance, get_processor_objective
from django_pain.utils import ProcessPaymentResult

LOGGER = getLogger(__name__)

PAYMENT_TYPE_CHOICES = (
    (PaymentType.TRANSFER, _("transfer")),
    (PaymentType.CARD_PAYMENT, _("card payment")),
)

PAYMENT_STATE_CHOICES = (
    (PaymentState.INITIALIZED, _("initialized")),
    (PaymentState.READY_TO_PROCESS, _("ready to process")),
    (PaymentState.PROCESSED, _("processed")),
    (PaymentState.DEFERRED, _("not identified")),
    (PaymentState.EXPORTED, _("exported")),
    (PaymentState.CANCELED, _("canceled")),
    (PaymentState.TO_REFUND, _("to refund")),
    (PaymentState.REFUNDED, _("refunded")),
)

PROCESSING_ERROR_CHOICES = (
    (PaymentProcessingError.DUPLICITY, _("Duplicate payment")),
    (PaymentProcessingError.INSUFFICIENT_AMOUNT, _("Received amount is lower than expected")),
    (PaymentProcessingError.EXCESSIVE_AMOUNT, _("Received amount is greater than expected")),
    (PaymentProcessingError.OVERDUE, _("Payment is overdue")),
    (PaymentProcessingError.MANUALLY_BROKEN, _("Payment was manually broken")),
    (PaymentProcessingError.TOO_OLD, _("Payment is older than 15 days, it can't be processed automatically")),
)


class BankAccount(models.Model):
    """Bank account."""

    account_number = models.TextField(unique=True, verbose_name=_("Account number"))
    account_name = models.TextField(blank=True, verbose_name=_("Account name"))
    currency = CurrencyField()

    class Meta:
        """Meta class."""

        verbose_name = _("Bank account")
        verbose_name_plural = _("Bank accounts")

    def __str__(self):
        """Return string representation of bank account."""
        return "{} {}".format(self.account_name, self.account_number)


class BankPayment(models.Model):
    """Bank payment."""

    identifier = models.TextField(verbose_name=_("Payment ID"))
    uuid = models.UUIDField(unique=True, editable=False, default=uuid.uuid4)
    payment_type = models.TextField(
        choices=PAYMENT_TYPE_CHOICES, default=PaymentType.TRANSFER, verbose_name=_("Payment type")
    )
    account = models.ForeignKey(BankAccount, on_delete=models.CASCADE, verbose_name=_("Destination account"))
    create_time = models.DateTimeField(auto_now_add=True, db_index=True, verbose_name=_("Create time"))
    transaction_date = models.DateField(null=True, db_index=True, verbose_name=_("Transaction date"))

    counter_account_number = models.TextField(blank=True, verbose_name=_("Counter account number"))
    counter_account_name = models.TextField(blank=True, verbose_name=_("Counter account name"))

    amount = MoneyField(max_digits=64, decimal_places=CURRENCY_PRECISION, verbose_name=_("Amount"))
    description = models.TextField(blank=True, verbose_name=_("Description"))
    state = models.TextField(
        choices=PAYMENT_STATE_CHOICES,
        default=PaymentState.READY_TO_PROCESS,
        db_index=True,
        verbose_name=_("Payment state"),
    )
    card_payment_state = models.TextField(blank=True, verbose_name=_("Card payment state"))

    processing_error = models.TextField(
        choices=PROCESSING_ERROR_CHOICES, null=True, blank=True, verbose_name=_("Automatic processing error")
    )

    # Payment symbols (specific for Czech Republic and Slovak Republic).
    constant_symbol = models.CharField(max_length=10, blank=True, verbose_name=_("Constant symbol"))
    variable_symbol = models.CharField(max_length=10, blank=True, verbose_name=_("Variable symbol"))
    specific_symbol = models.CharField(max_length=10, blank=True, verbose_name=_("Specific symbol"))

    processor = models.TextField(verbose_name=_("Processor"), blank=True)
    card_handler = models.TextField(verbose_name=_("Card handler"), blank=True)

    # CSOB custom payment specific fields
    email = models.EmailField(null=True, blank=True, verbose_name=_("Customer email"))
    custom_expiry = models.DateTimeField(null=True, blank=True, verbose_name=_("Payment expiry"))
    customer_code = models.CharField(max_length=10, blank=True, verbose_name=_("Customer code"))
    email_sent = models.DateTimeField(null=True, blank=True, verbose_name=_("Email sending time"))

    class Meta:
        """Model Meta class."""

        unique_together = ("identifier", "account")
        verbose_name = _("Bank payment")
        verbose_name_plural = _("Bank payments")

        constraints = [
            CheckConstraint(
                check=(
                    Q(payment_type=PaymentType.TRANSFER)
                    | Q(payment_type=PaymentType.CARD_PAYMENT, counter_account_number__exact="")
                ),
                name="payment_counter_account_only_for_transfer",
            )
        ]

    def __str__(self):
        """Return string representation of bank payment."""
        return self.identifier

    def clean(self):
        """Check whether payment currency is the same as currency of related bank account."""
        if self.account.currency != self.amount.currency.code:
            raise ValidationError(
                "Bank payment {} is in different currency ({}) than bank account {} ({}).".format(
                    self.identifier, self.amount.currency.code, self.account, self.account.currency
                )
            )
        super().clean()

    @property
    def advance_invoice(self):
        """Return advance invoice if it exists."""
        invoices = self.invoice_refs.filter(invoice_type=InvoiceType.ADVANCE)
        return invoices.first()

    @property
    def objective(self):
        """Return processed payment objective."""
        if self.processor:
            return get_processor_objective(self.processor)
        else:
            return ""

    objective.fget.short_description = _("Objective")  # type: ignore

    @classmethod
    def objective_choices(self):
        """Return payment processor default objectives choices."""
        choices = BLANK_CHOICE_DASH.copy()
        for proc_name in SETTINGS.processors:
            proc = get_processor_instance(proc_name)
            choices.append((proc_name, proc.default_objective))
        return choices

    def save_processing_result(self, result: ProcessPaymentResult) -> None:
        """Update payment state based on result from payment processor.

        This is utility function for uniform processing of result from payment processor (which doesn't prevent
        different handling if required).

        Note: Payment is expected in READY_TO_PROCESS or DEFERRED state.
        """
        if result.result == ProcessPaymentResultType.PROCESSED:
            self.state = PaymentState.PROCESSED
        elif result.result == ProcessPaymentResultType.TO_REFUND:
            LOGGER.info("Saving payment %s as TO_REFUND.", self.uuid)
            self.state = PaymentState.TO_REFUND
        else:
            LOGGER.info("Saving payment %s as DEFERRED with error %s.", self.uuid, result.error)
            self.state = PaymentState.DEFERRED
        self.processing_error = result.error
        self.save()


class PaymentImportHistory(models.Model):
    """Record of payment imports."""

    separator = ";"

    origin = models.TextField(verbose_name=_("Origin"), help_text="Key of PAIN_DOWNLOADERS setting.")
    start_datetime = models.DateTimeField(
        auto_now_add=True, db_index=True, verbose_name=_("Import start time"), help_text="Import start time."
    )
    _filenames = models.TextField(
        null=True,
        blank=True,
        verbose_name=_("File names"),
        help_text="Names of the files the data were imported from if known.",
    )
    errors = models.PositiveSmallIntegerField(
        null=True, blank=True, verbose_name=_("Errors"), help_text="Number of payments skipped due to an error."
    )
    finished = models.BooleanField(
        default=False,
        verbose_name=_("Finished"),
        help_text="Indicates whether the import command finished without raising an error.",
    )

    class Meta:
        """Model Meta class."""

        verbose_name = _("Payment Import History")
        verbose_name_plural = _("Payment Import History")

    @property
    def success(self) -> bool:
        """Indicate successful import, i.e. the command finished and there were no errors."""
        return self.finished and self.errors == 0

    success.fget.short_description = _("Success")  # type: ignore

    @property
    def filenames(self) -> Tuple[str, ...]:
        """Wrap _filenames field to return iterator."""
        if self._filenames is None:
            return ()
        else:
            return tuple(self._filenames.split(self.separator))

    filenames.fget.short_description = _("File names")  # type: ignore

    def add_filename(self, filename: str) -> None:
        """Add value to _filenames field so it stays consistent."""
        if self.separator in filename:
            raise ValueError("Character {} not allowed in filenames.".format(self.separator))
        self._filenames = self.separator.join(self.filenames + (filename,))

    def __str__(self) -> str:
        """Return string representation of an import record."""
        return "{} {}".format(self.origin, self.start_datetime)

==> django_pain-4.1.0-py3-none-any.whl: django_pain/models/bank.py <==
#
# Copyright (C) 2018-2025  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""Payments and invoices models."""

import uuid

from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import BLANK_CHOICE_DASH, CheckConstraint, Q
from django.utils.translation import gettext_lazy as _
from djmoney.models.fields import CurrencyField, MoneyField

from django_pain.constants import (
    CURRENCY_PRECISION,
    InvoiceType,
    PaymentProcessingError,
    PaymentState,
    PaymentType,
    ProcessPaymentResultType,
)
from django_pain.logging import getLogger
from django_pain.settings import SETTINGS, get_processor_instance, get_processor_objective
from django_pain.utils import ProcessPaymentResult

LOGGER = getLogger(__name__)

PAYMENT_TYPE_CHOICES = (
    (PaymentType.TRANSFER, _("transfer")),
    (PaymentType.CARD_PAYMENT, _("card payment")),
)

PAYMENT_STATE_CHOICES = (
    (PaymentState.INITIALIZED, _("initialized")),
    (PaymentState.READY_TO_PROCESS, _("ready to process")),
    (PaymentState.PROCESSED, _("processed")),
    (PaymentState.DEFERRED, _("not identified")),
    (PaymentState.EXPORTED, _("exported")),
    (PaymentState.CANCELED, _("canceled")),
    (PaymentState.TO_REFUND, _("to refund")),
    (PaymentState.REFUNDED, _("refunded")),
)

PROCESSING_ERROR_CHOICES = (
    (PaymentProcessingError.DUPLICITY, _("Duplicate payment")),
    (PaymentProcessingError.INSUFFICIENT_AMOUNT, _("Received amount is lower than expected")),
    (PaymentProcessingError.EXCESSIVE_AMOUNT, _("Received amount is greater than expected")),
    (PaymentProcessingError.OVERDUE, _("Payment is overdue")),
    (PaymentProcessingError.MANUALLY_BROKEN, _("Payment was manually broken")),
    (PaymentProcessingError.TOO_OLD, _("Payment is older than 15 days, it can't be processed automatically")),
)


class BankAccount(models.Model):
    """Bank account."""

    uuid = models.UUIDField(unique=True, editable=False, default=uuid.uuid4)
    account_number = models.TextField(unique=True, verbose_name=_("Account number"))
    account_name = models.TextField(blank=True, verbose_name=_("Account name"))
    currency = CurrencyField()

    class Meta:
        """Meta class."""

        verbose_name = _("Bank account")
        verbose_name_plural = _("Bank accounts")

    def __str__(self):
        """Return string representation of bank account."""
        return "{} {}".format(self.account_name, self.account_number)


class BankPayment(models.Model):
    """Bank payment."""

    identifier = models.TextField(verbose_name=_("Payment ID"))
    uuid = models.UUIDField(unique=True, editable=False, default=uuid.uuid4)
    payment_type = models.TextField(
        choices=PAYMENT_TYPE_CHOICES, default=PaymentType.TRANSFER, verbose_name=_("Payment type")
    )
    account = models.ForeignKey(BankAccount, on_delete=models.CASCADE, verbose_name=_("Destination account"))
    create_time = models.DateTimeField(auto_now_add=True, db_index=True, verbose_name=_("Create time"))
    transaction_date = models.DateField(null=True, db_index=True, verbose_name=_("Transaction date"))

    counter_account_number = models.TextField(blank=True, verbose_name=_("Counter account number"))
    counter_account_name = models.TextField(blank=True, verbose_name=_("Counter account name"))

    amount = MoneyField(max_digits=64, decimal_places=CURRENCY_PRECISION, verbose_name=_("Amount"))
    description = models.TextField(blank=True, verbose_name=_("Description"))
    state = models.TextField(
        choices=PAYMENT_STATE_CHOICES,
        default=PaymentState.READY_TO_PROCESS,
        db_index=True,
        verbose_name=_("Payment state"),
    )
    card_payment_state = models.TextField(blank=True, verbose_name=_("Card payment state"))

    processing_error = models.TextField(
        choices=PROCESSING_ERROR_CHOICES, null=True, blank=True, verbose_name=_("Automatic processing error")
    )

    # Payment symbols (specific for Czech Republic and Slovak Republic).
    constant_symbol = models.CharField(max_length=10, blank=True, verbose_name=_("Constant symbol"))
    raw_variable_symbol = models.CharField(
        db_column="variable_symbol", max_length=10, blank=True, verbose_name=_("Variable symbol")
    )
    specific_symbol = models.CharField(max_length=10, blank=True, verbose_name=_("Specific symbol"))

    processor = models.TextField(verbose_name=_("Processor"), blank=True)
    card_handler = models.TextField(verbose_name=_("Card handler"), blank=True)

    # CSOB custom payment specific fields
    email = models.EmailField(null=True, blank=True, verbose_name=_("Customer email"))
    custom_expiry = models.DateTimeField(null=True, blank=True, verbose_name=_("Payment expiry"))
    customer_code = models.CharField(max_length=10, blank=True, verbose_name=_("Customer code"))
    email_sent = models.DateTimeField(null=True, blank=True, verbose_name=_("Email sending time"))

    class Meta:
        """Model Meta class."""

        unique_together = ("identifier", "account")
        verbose_name = _("Bank payment")
        verbose_name_plural = _("Bank payments")

        constraints = [
            CheckConstraint(
                check=(
                    Q(payment_type=PaymentType.TRANSFER)
                    | Q(payment_type=PaymentType.CARD_PAYMENT, counter_account_number__exact="")
                ),
                name="payment_counter_account_only_for_transfer",
            )
        ]

    def __str__(self):
        """Return string representation of bank payment."""
        return self.identifier

    def clean(self):
        """Check whether payment currency is the same as currency of related bank account."""
        if self.account.currency != self.amount.currency.code:
            raise ValidationError(
                "Bank payment {} is in different currency ({}) than bank account {} ({}).".format(
                    self.identifier, self.amount.currency.code, self.account, self.account.currency
                )
            )
        super().clean()

    @property
    def advance_invoice(self):
        """Return advance invoice if it exists."""
        invoices = self.invoice_refs.filter(invoice_type=InvoiceType.ADVANCE)
        return invoices.first()

    @property
    def objective(self):
        """Return processed payment objective."""
        if self.processor:
            return get_processor_objective(self.processor)
        else:
            return ""

    objective.fget.short_description = _("Objective")  # type: ignore

    @classmethod
    def objective_choices(self):
        """Return payment processor default objectives choices."""
        choices = BLANK_CHOICE_DASH.copy()
        for proc_name in SETTINGS.processors:
            proc = get_processor_instance(proc_name)
            choices.append((proc_name, proc.default_objective))
        return choices

    def save_processing_result(self, result: ProcessPaymentResult) -> None:
        """Update payment state based on result from payment processor.

        This is utility function for uniform processing of result from payment processor (which doesn't prevent
        different handling if required).

        Note: Payment is expected in READY_TO_PROCESS or DEFERRED state.
        """
        if result.result == ProcessPaymentResultType.PROCESSED:
            self.state = PaymentState.PROCESSED
        elif result.result == ProcessPaymentResultType.TO_REFUND:
            LOGGER.info("Saving payment %s as TO_REFUND.", self.uuid)
            self.state = PaymentState.TO_REFUND
        else:
            LOGGER.info("Saving payment %s as DEFERRED with error %s.", self.uuid, result.error)
            self.state = PaymentState.DEFERRED
        self.processing_error = result.error
        self.save()

    @property
    def variable_symbol(self) -> str:
        """Return normalized variable symbol.

        Leading zeroes are possibly trimmed (respect settings).
        """
        return self.raw_variable_symbol.lstrip("0") if SETTINGS.trim_varsym else self.raw_variable_symbol

    @variable_symbol.setter
    def variable_symbol(self, value: str) -> None:
        """Set variable symbol value."""
        self.raw_variable_symbol = value


class PaymentImportHistory(models.Model):
    """Record of payment imports."""

    separator = ";"

    uuid = models.UUIDField(unique=True, editable=False, default=uuid.uuid4)
    origin = models.TextField(verbose_name=_("Origin"), help_text="Key of PAIN_DOWNLOADERS setting.")
    start_datetime = models.DateTimeField(
        auto_now_add=True, db_index=True, verbose_name=_("Import start time"), help_text="Import start time."
    )
    _filenames = models.TextField(
        null=True,
        blank=True,
        verbose_name=_("File names"),
        help_text="Names of the files the data were imported from if known.",
    )
    errors = models.PositiveSmallIntegerField(
        null=True, blank=True, verbose_name=_("Errors"), help_text="Number of payments skipped due to an error."
    )
    finished = models.BooleanField(
        default=False,
        verbose_name=_("Finished"),
        help_text="Indicates whether the import command finished without raising an error.",
    )

    class Meta:
        """Model Meta class."""

        verbose_name = _("Payment Import History")
        verbose_name_plural = _("Payment Import History")

    @property
    def success(self) -> bool:
        """Indicate successful import, i.e. the command finished and there were no errors."""
        return self.finished and self.errors == 0

    success.fget.short_description = _("Success")  # type: ignore

    @property
    def filenames(self) -> tuple[str, ...]:
        """Wrap _filenames field to return iterator."""
        if self._filenames is None:
            return ()
        else:
            return tuple(self._filenames.split(self.separator))

    filenames.fget.short_description = _("File names")  # type: ignore

    def add_filename(self, filename: str) -> None:
        """Add value to _filenames field so it stays consistent."""
        if self.separator in filename:
            raise ValueError("Character {} not allowed in filenames.".format(self.separator))
        self._filenames = self.separator.join(self.filenames + (filename,))

    def __str__(self) -> str:
        """Return string representation of an import record."""
        return "{} {}".format(self.origin, self.start_datetime)
